/*******************************************************************************
 * Copyright (c) 2013-2015 Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser Public License v3
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl.html
 *
 * Contributors:
 *     Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza - initial API and implementation
 ******************************************************************************/
package com.net2plan.examples.onlineSim.timeVaryingTraffic;

import com.net2plan.interfaces.networkDesign.NetPlan;
import com.net2plan.interfaces.simulation.IEventProcessor;
import com.net2plan.interfaces.simulation.SimAction;
import com.net2plan.interfaces.simulation.SimEvent;
import com.net2plan.libraries.CandidatePathList;
import com.net2plan.libraries.WDMUtils;
import com.net2plan.utils.LongUtils;
import com.net2plan.utils.Pair;
import com.net2plan.utils.StringUtils;
import com.net2plan.utils.Triple;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * <p>At each time interval, this algorithm adjusts dynamically the number of lightpaths 
 * for each demand with respect to the ones in the previous period.</p>
 * 
 * <p>In case that new lightpaths are required, it performs an RWA using first-fit 
 * (wavelength continuity is enforced) over a set of precomputed <tt>k</tt>-shortest paths 
 * (maximum path length in km equal to <tt>maxLightpathLengthInKm</tt>). These paths are 
 * evaluated in order. Once a feasible RWA is found, then the lightpath 
 * is allocated, and no other candidate solution is checked.</p>
 * 
 * <p>Let <i>currentNumLps<sub>d</sub></i> be the number of lightpaths which are active for 
 * demand <i>d &isin; D</i> in the current time, <i>h<sub>d</sub></i> the new traffic volume in 
 * Gbps for demand <i>d &isin; D</i>, and <tt>binaryRatePerChannel_Gbps</tt> the binary 
 * rate per lightpath. Then, the number of lightpaths for demand <i>d &isin; D</i>, 
 * denoted as <i>numLps<sub>d</sub></i>, for the new period is computed as follows:</p>
 * 
 * <p><i>numLps<sub>d</sub> = ceil(h<sub>d</sub>/binaryRatePerChannel_Gbps)</i></p>
 * 
 * <p>Then, we adjust the number of lightpaths as follows:
 * <ul>
 *  <li>If <i>numLps<sub>d</sub> - currentNumLps<sub>d</sub> > 0</i>, add <i>numLps<sub>d</sub> - currentNumLps<sub>d</sub></i>
 *  lightpaths using the pre-computed k-shortest paths and first-fit wavelength assignment</li>
 *  <li>If <i>numLps<sub>d</sub> - currentNumLps<sub>d</sub> < 0</i>, remove the latest <i>numLps<sub>d</sub> - currentNumLps<sub>d</sub></i>
 *  lightpaths from demand</li>
 *  <li>Otherwise, do nothing</li>
 * </ul></p>
 *
 * @author Pablo Pavon-Marino, Jose-Luis Izquierdo-Zaragoza
 * @version 1.1, May 2015
 */
public class TVSim_AA_WDM_basicTrafficAllocationAlgorithm extends IEventProcessor
{
	private Map<Long, Pair<Long, Long>> fiberMap;
	
	private double binaryRatePerChannel_Gbps;
	private CandidatePathList cpl;
	private Map<Long, Integer> w_f;
	private List<Set<Long>> wavelengthFiberOccupancy;

	@Override
	public String getDescription()
	{
		String newLine = StringUtils.getLineSeparator();

		StringBuilder description = new StringBuilder();
		description.append("This algorithm adjusts dynamically the number of "
				+ "lightpaths for each demand with respect to the ones in the "
				+ "previous period.");
		description.append(newLine).append(newLine);
		description.append("At each time interval, this algorithm adjusts "
				+ "dynamically the number of lightpaths for each demand with "
				+ "respect to the ones in the previous period.");
		description.append(newLine).append(newLine);
		description.append("In case that new lightpaths are required, it "
				+ "performs an RWA using first-fit (wavelength continuity "
				+ "is enforced) over a set of precomputed k-shortest paths "
				+ "(maximum path length in km equal to maxLightpathLengthInKm). "
				+ "These paths are evaluated in order. Once a feasible RWA is found, "
				+ "then the lightpath is allocated, and no other candidate solution is checked.");

		return description.toString();
	}

	@Override
	public List<Triple<String, String, String>> getParameters()
	{
		List<Triple<String, String, String>> algorithmParameters = new LinkedList<Triple<String, String, String>>();
		algorithmParameters.add(Triple.of("binaryRatePerChannel_Gbps", "10", "Binary rate of all the lightpaths"));
		algorithmParameters.add(Triple.of("maxLightpathLengthInKm", "5000", "Maximum allowed lightpath length in km"));
		algorithmParameters.add(Triple.of("k", "3", "Maximum number of candidate paths per demand"));

		return algorithmParameters;
	}

	@Override
	public void initialize(NetPlan initialNetPlan, NetPlan currentNetPlan, Map<String, String> algorithmParameters, Map<String, String> simulationParameters, Map<String, String> net2planParameters)
	{
		WDMUtils.checkConsistency(currentNetPlan);

		int k = Integer.parseInt(algorithmParameters.get("k"));
		double maxLightpathLengthInKm = Double.parseDouble(algorithmParameters.get("maxLightpathLengthInKm"));

		binaryRatePerChannel_Gbps = Double.parseDouble(algorithmParameters.get("binaryRatePerChannel_Gbps"));
		cpl = new CandidatePathList(currentNetPlan, currentNetPlan.getLinkLengthInKmMap(), "maxLengthInKm", Double.toString(maxLightpathLengthInKm), "K", Integer.toString(k));
		fiberMap = currentNetPlan.getLinkMap();
		w_f = WDMUtils.getFiberNumWavelengthsMap(currentNetPlan);
		wavelengthFiberOccupancy = WDMUtils.getWavelengthFiberOccupancy(currentNetPlan);
	}

	@Override
	public void processEvent(NetPlan currentNetPlan, SimEvent event)
	{
		List<SimAction> actions = event.getEventActionList();
		for(SimAction action : actions)
		{
			if (action.getActionType() == SimAction.ActionType.DEMAND_MODIFIED)
			{
				long lpDemandId = action.getDemandModified_demandId();
				int numLps = (int) Math.ceil(currentNetPlan.getDemandOfferedTraffic(lpDemandId) / binaryRatePerChannel_Gbps);
				long[] currentLpIds = LongUtils.toArray(currentNetPlan.getDemandRoutes(lpDemandId));

				if (numLps > currentLpIds.length)
				{
					/* add new lightpaths */
					int numLpsToBeSetUp = numLps - currentLpIds.length;

					for (int r = 0; r < numLpsToBeSetUp; r++)
					{
						for (long pathId : cpl.getDemandPaths(lpDemandId))
						{
							List<Long> seqFibers_thisLp = cpl.getPathSequenceOfLinks(pathId);
							int[] seqWavelengths_thisLp = WDMUtils.WA_firstFit(fiberMap, seqFibers_thisLp, w_f, wavelengthFiberOccupancy);
							if (seqWavelengths_thisLp.length == 0) continue;

							/* add this route, updating wavelength occupation */
							long lpId = currentNetPlan.addRoute(lpDemandId, binaryRatePerChannel_Gbps, seqFibers_thisLp, null);
							WDMUtils.allocateResources(seqFibers_thisLp, seqWavelengths_thisLp, wavelengthFiberOccupancy);
							WDMUtils.setLightpathSeqWavelengths(currentNetPlan, lpId, seqWavelengths_thisLp);
							break;
						}
					}
				}
				else if (numLps < currentLpIds.length)
				{
					/* eliminate last lightpaths, if I have to reduce the amount of lightpaths */
					int numLpsToBeTornDown = currentLpIds.length - numLps;

					for (int r = 0; r < numLpsToBeTornDown; r++)
					{
						long lpId = currentLpIds[currentLpIds.length - 1 - r];

						List<Long> seqFibers_thisLp = currentNetPlan.getRouteSequenceOfLinks(lpId);
						int[] seqWavelengths_thisLp = WDMUtils.getLightpathSeqWavelengths(currentNetPlan, lpId);
						WDMUtils.releaseResources(seqFibers_thisLp, seqWavelengths_thisLp, wavelengthFiberOccupancy);
					}
				}
			}
		}
	}
}
